#! /usr/bin/env python
# -*- coding: utf-8 -*-
#########################################################################
# Copyright (C) 2011 Niels Bassler, bassler@phys.au.dk
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
# Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, see <http://www.gnu.org/licenses/>.
#
# rcshield.py
#
# Script which eased the submission of a shield job to a
# condor cluster.
#
# NB: Niels Bassler, bassler@phys.au.dk,
# RH: Rochus Herrmann, rhe@phys.au.dk
# GPL License applies.
#
# Note that the statistics is increased according to the
# amount of nodes required.
#
# Change log:
#   - 19.10.2011, NB Fork from rcfluka.py
#   - 09.09.2014, NB Update input files.
#
# TODO:
#   - Consequent checking existence of files etc.
#   - better  complete clean "-C"
#   - standard universe.
#   - test timeouts
#   - reduce long temporary filenames
#   - also remove ranfluka* with -c -C

import os, re, sys
import shutil # for fast file copy
import time, random, glob
from subprocess import Popen
from optparse import OptionParser

version = "1.0"

def condor_ids():  # returns list of running processes
    active_ids = {}
    pipe = os.popen("condor_q -format \"%f\\n\" clusterid",'r')
    output = pipe.readlines()
    pipe.close()
    for lines in output:
        a_id = float(lines)
        active_ids[a_id] = 1
    # this is a list of active keys now:
    return(active_ids.keys())

def cleanup():
    # cleanup, improve this
    print sys.argv[0] + ": ---- Cleanup  -------------------------------"
    for filename in glob.glob('_rcshield*') :
        os.remove( filename )


input_files = ["geo.dat","beam.dat","mat.dat","detect.dat"] # mandatory input files
nr_jobs = 1

seed = 0.0    # init this value for deterministic runs
# parse options:
args = sys.argv[1:]
if args.__len__() < 1:
    print ""
    print "rcshield.py Version", version
    print ""
    print "Error: Number of runs not stated."
    print ""
    print "Usage example: run input on 5 nodes"
    print "rcshield.py -M5 [directory]"
    print ""
    print "Type rcshield.py -h for complete option list."
    print ""
    sys.exit(0)

parser = OptionParser()
parser.add_option("-f", "--file", dest="FILE",
                  help="additional optional files, such as ripple filters, external beam sources or external stopping power files.", metavar="FILE")
parser.add_option("-M", "--MACHINES", dest="max",
                  default=1,help="Number of jobs to submit.", metavar="int")
parser.add_option("-N", "--NUM", dest="num", default=0,
                  help="Start at job number N. Default is 0.", metavar="int")
parser.add_option("-a", "--nowait", action="store_true", dest="fast",
                  default=False, help="Do not wait for finishing file transfer.")
parser.add_option("-n", "--nice", action="store_true", dest="nice",
                  default=False, help="Submit as nice job.")
parser.add_option("-c", "--clean", action="store_true", dest="clean",
                  default=False, help="Remove temporary files. Use carefully.")
parser.add_option("-C", "--bigclean", action="store_true", dest="bigclean",
                  default=False, help="Remove almost all files, including data. Dangerous!")
parser.add_option("-t", "--test", action="store_true", dest="test",
                  default=False, help="Test, do not submit to condor.")
parser.add_option("-S", "--standard", action="store_true", dest="standard",
                  default=False, help="Submit to standard universe.")
parser.add_option("-o", "--optioncondor", dest="OPTC",
                  help="Additional optional string to be included in condor submit file.", metavar="string")
parser.add_option("-i", "--includecondor", dest="INCC",
                  help="Append contents of FILE in condor submit file.", metavar="FILE")
parser.add_option("-m", "--mail", dest="mail",
		  help="Sends report to given e-mail address when job is finished", metavar="string")
parser.add_option("-x", "--single", action="store_true", dest="SINGLE",
                  default=False, help="Submit only once and exit after submission. This overrides -M and --mail option. Used to avoid funky node suffixes.")
parser.add_option("-q", "--quiet",
                  action="store_false", dest="verbose", default=True,
                  help="don't print status messages to stdout")
(options, args) = parser.parse_args()

# get current working directory
_cwd = os.getcwd()
# check if args[0] exist

# if sepcified, change to directory with input data
base_path = ""
if len(args) > 0:
    if os.path.isdir(args[0]):
        base_path = args[0]
        os.chdir(base_path)

if options.bigclean == True:
    print "Removing all generated files."
    for filename in glob.glob('_rcshield*') :
        os.remove( filename )
    for filename in glob.glob('*[0-9][0-9][0-9][0-9]') :
        os.remove( filename )
    for filename in glob.glob('for[0-9][0-9][0-9]') :
        os.remove( filename )
    os.remove('fort.24') # no idea where this comes from
    sys.exit()


# hash table for condor_submissions:
sub_list = {}

nr_jobs_start = int(options.num)
nr_jobs = int(options.max)

#print "Inputfilename = "
#print input_filename
#print type(nr_jobs), nr_jobs

if options.SINGLE == True:
    nr_jobs = 1
print "Preparing ", nr_jobs, " condor submissions..."

# increase if you want. This is just for protection of typing errors etc.
if nr_jobs > 9999:
    print sys.argv[0] +" Error: ---- No more than 9999 submissions supported. Exiting."
    sys.exit()

# generate new file names for each node.

if options.INCC != None:
    if os.path.isfile(options.INCC) is False:
        print sys.argv[0] +" Error: ---- Could not find file "+ options.INCC + "Exiting."
    	sys.exit()
    # read files
    inc_condor = open(options.INCC, 'r').readlines()


for i in range(nr_jobs) :
    i += nr_jobs_start
    if options.standard == False :
        condor_universe = "vanilla"
    else:
        condor_universe = "standard"
    condor_filename = "_rcshield_submit%04d.cdr" % i
    condor_stdout = "_rcshield_stdout_%04d.txt" % i
    condor_log = "_rcshield.log"


    #######################  generate condor script
    #    if nr_jobs == 0:
    file = open(condor_filename, "w")
    condor_string =  "##################################################\n"
    condor_string += "# rcshield.py autogenerated condor job description.\n"
    condor_string += "Executable      = "  + "/usr/export/bin/shieldhit" + "\n"
    if options.SINGLE == False:
        condor_string += "Arguments       = "  + "-N" +str(i)+ "\n"
    condor_string += "Universe        = " + condor_universe + "\n"
    condor_string += "Output          = " + condor_stdout + "\n"
    # just on abacus
    # TODO: require linux (instead).
    condor_string += "Requirements = Arch == \"X86_64\" && Memory >= 1800         \n"
    #
    if options.OPTC != None:
        condor_string += options.OPTC + "\n"
    # nice job?
    if options.nice == True:
	condor_string += "nice_user = True\n"

    # add input condor file bla bla here
    if options.INCC != None:
        for lines in inc_condor:
            condor_string += lines

    # rh
    # condor just notifies if a job experiences an error, not at every
    # successfully completed job
    condor_string += "Notification = Error \n"
    # hr

    condor_string += "Log             = " + condor_log + "\n"
    condor_string += "should_transfer_files  = YES\n"
    condor_string += "when_to_transfer_output = ON_EXIT\n"
    condor_string += "transfer_input_files = "
    condor_string += (", ").join(input_files)
    if options.FILE != None:
        for fff in options.FILE.split(","):
            condor_string += ", " + fff
    condor_string +="\n"
    condor_string += "Queue\n"
##     else:
##         condor_string += "Executable = " + shell_filename + "\n"
##         condor_string += "transfer_input_files = " temp_filename + ".inp"
##         for fff in options.SOURCE-split(","):
##             condor_string += " " + fff
##             condor_string +="\n"
##             condor_string += "Queue\n"
    file.write(condor_string)
    file.close()

    # execute next line with popen
    exec_string = "condor_submit "+condor_filename+ "\n"
#    print exec_string

    if options.test==False :
        # submission process is unparallelized
        sss = os.popen(exec_string)
        # grab stdout
        sss_lines = sss.readlines()
        sss.close() # is this correct?
        sss_id = float(sss_lines[-1].split()[5]) # needs float because of .
        print "Submitted ", condor_filename, "with Condor id: ", sss_id
        sys.stdout.flush() # flush output for nohup
        sub_list[sss_id] = 1 # 1 for running


    #    p.append(Popen(exec_string, shell=True))
    # aparantly there is some problem, if submitted too fast?
    #    time.sleep(1)

if options.SINGLE == True:
    #cleanup no good idea, since it will remove open files which is still used by condor.
    #if options.clean == True:
    #    time.sleep(5)
    #    cleanup()
    print " ---- Single condor job is now running. Exit. ----- "
    sys.exit()

print "-------------- Now waiting for completed jobs ------------------"

if options.test == False:
    finito = False
    timeout = 0
    # spaghetthi time!
    while finito is False:
        timeout += 1
        time.sleep(15) # check every 15 seconds.
        # update sub_list with processes started by this invokation
        # mark all as done
        finito = True
        for b in sub_list.keys():
            if sub_list[b] == 1:
                sub_list[b] = 2
            else:
                sub_list[b] = 0
        # get current list of active processes:
        id_list = condor_ids()
        # and remark list
        for b in id_list:
            if (b in sub_list):
                sub_list[b] = 1
                finito = False
        # now all keys with 2 recently completed
        for b in sub_list.keys():
            if sub_list[b] == 2:
                print "rcshield submission ",b," terminated."
                sys.stdout.flush()
                sub_list[b] = 0

        if timeout > 200000:
            print "rcshield.py TIMEOUT Error!"
            print "The submitted job took more than 34 days to complete and we have run out of cake."
            finito = True

        # or stop if no condor jobs are running
        if id_list == []:
            finito = True

# then start looping over all data stuff


##for i in range(nr_jobs):
    # old version, get pids
    # os.waitpid(p[i].pid,0)
##    temp_filename     = input_filename + "_node%04d" % i
##    temp_filename_out = input_filename + "_node%04d001" % i + ".out"
##    temp_filename_inp = input_filename + "_node%04d" % i + ".inp"
##    print temp_filename_out
##    print input_filename
##    if options.test == False:
##
##        # wait for file transfer to complete
##        timeout = 0 # timout
##        while timeout < 600:
##            if os.path.isfile(temp_filename_out) is True:
##                timeout = 600
##            timeout += 1
##            time.sleep(1)
##        os.remove(temp_filename_inp)


print sys.argv[0] + ": ---- Done calculations -------------------------------"
if options.fast == False:
    print sys.argv[0] + "  We will now wait 30 seconds for condor to complete"
    print sys.argv[0] + "  the file transfer of your data."
    time.sleep(30)

#print sys.argv[0] + ": ---- Copy files        -------------------------------"
# sleep a lot of seconds, for condor to complete file transfer
# race condition could be encountered here.


# after execution, rename all files to normal terminology
#current_dirlist = os.listdir(os.getcwd())

#foo_node01001_fort.XX -> foo001_fort.XX
#foo_node02001_fort.XX -> foo002_fort.XX
#foo_node03001_fort.XX -> foo003_fort.XX
#etc

#for ii in range(nr_jobs) :
#    i = ii + start_job
#    pattern_nodeid = "_node%04d" %i + "001"
#    pattern_nodeid_new = "0%04d" %i
#
#    #attern = "^"+input_filename + "_node%04d" % i + "001"
#    pattern =     input_filename + "_node%04d" % i + "001"
#
#    allowed_name = re.compile(pattern).match
#    print pattern
#    for fname in current_dirlist:
#        if allowed_name(fname):
#            new_filename = fname.replace(pattern_nodeid, pattern_nodeid_new)
#            print "Renameing", fname, "->", new_filename
#            new_filname = os.rename(fname, new_filename)


# send mail
if options.mail != None:
    print sys.argv[0] + ": ---- Send Report to: "+options.mail+" --------------------"
    os.system("mail -s \'"+base_path +": job_finished\' -c \'\' "+ options.mail+ " <_rcshield.log")

if options.clean == True:
    cleanup()

# go home to directory where this command was invoked from
# os.chdir(_cwd)
# Not needed?

print sys.argv[0] + ": ---- Finished  -------------------------------"
